/**********************************************************\
| Michael Hamilton (maxmike@gmail.com) www.mutedvision.net |
|----------------------------------------------------------|
| Primitive drawing shapes go here.  Common notation for   |
| lists of points is start at the top left corner and go   |
| clockwise (so 1 = top left, 2 = top right, 3 = bot right |
| 4 = bot left)                                            |
\**********************************************************/

#ifndef _DRAWSHAPES_H_
#define _DRAWSHAPES_H_

#include <stdio.h>
#include <iostream>
#include <algorithm>
#include <string>
#include <vector>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include "boost/lexical_cast.hpp"
#include "Render/textures.h"
#include "Render/points.h"
#include "Render/render.h"
#include "Utility/package.h"

namespace M2Rend {
   const double PIE = 3.14159265358979323846;

   class BoxAABB{
   public:
      BoxAABB(){initialize(Point());}
      BoxAABB(const Point &startPoint){initialize(startPoint);}
      BoxAABB(const Point &startPoint, const Point &endPoint){initialize(startPoint, endPoint);}

      void initialize(const Point &startPoint);
      void initialize(const BoxAABB &startBox);
      void initialize(const Point &startPoint, const Point &endPoint);
      void expandWith(const Point &comparePoint);
      void expandWith(const BoxAABB &compareBox);
      bool pointContained(const Point &comparePoint);
      bool pointContained(double x, double y, double z);
      bool pointContainedXY(const Point &comparePoint);
      bool pointContainedXY(double x, double y);
      Point minPoint, maxPoint;
   };

   class PointVolume{
   public:
      void addPoint(const Point &newPoint);
      bool pointContained(double x, double y);
      bool pointContained(const Point &comparePoint);
      
      bool volumeCollision(PointVolume &compareVolume, Draw2D* renderer);
      Point getCenter();
      BoxAABB getAABB();

      std::vector<Point> points;
   private:
      //get angle within proper range between two points (-pi to +pi)
      double getAngle(double x1, double y1, double x2, double y2);
   };

   class DrawShape{
   public:
      DrawShape(){renderer=0;myParent=0;hasTexture = false;depthOverride = false;}
      DrawShape(Draw2D* context){renderer=context;myParent=0;hasTexture = false;depthOverride = false;}
      virtual ~DrawShape(){}

      virtual void setRenderer(Draw2D* context){renderer = context;}

      //For composite collections
      virtual DrawShape* add(DrawShape* childItem, std::string childId){return false;}
      virtual bool remove(DrawShape* childItem, bool deleteOnRemove = true){return false;}
      virtual bool remove(std::string childId, bool deleteOnRemove = true){return false;}
      virtual DrawShape* getChild(std::string childId){return 0;}
      virtual void draw() = 0;
      virtual double getDepth();

      //Axis Aligned Bounding Box
      virtual BoxAABB getAABB();
      //Get all points
      virtual PointVolume getWorldPoints();
      virtual Point getWorldPoint(Point local);
      //For command type alerts
      void setParent(DrawShape* parentItem);

      void scale(double newScale);
      void scale(double x, double y, double z);
      void scale(Point scaleValue);
      void incrementScale(double newScale);
      void incrementScale(double x, double y, double z);
      void incrementScale(Point scaleValue);

      //Rotation is degrees around each axis from 0 to 360
      void setRotate(double x, double y, double z);
      void setRotate(Point rotation){rotateTo = rotation;}
      void incrementRotate(double x, double y, double z);
      void incrementRotate(Point rotation){rotateTo+= rotation;}
      void setRotateOrigin(double x, double y, double z);
      void setRotateOrigin(Point origin){rotateOrigin = origin;}

      void translate(double x, double y, double z);
      void translate(Point translation){translateTo+= translation;}
      void placeAt(double x, double y, double z);
      void placeAt(Point placement){translateTo = placement;}

      //By default the sort depth is calculated by averaging the z values of all points
      //setSortDepth manually overrides this calculation.  unsetSortDepth removes this override.
      void setSortDepth(double newDepth){depthOverride = true; overrideDepthValue = newDepth;}
      void unsetSortDepth(){depthOverride = false;}

      Point getLocation();
      Point getRotation();
      Point getScale();

      virtual Point getRelativeLocation();
      virtual Point getRelativeRotation();
      virtual Point getRelativeScale();
      
      virtual void setColor(float R, float G, float B, float A = 1.0);
      virtual void setColor(Color newColor);

      void setTexture(GLuint *textureId);
      void removeTexture();

      //Used to determine sorting based on getDepth
      bool operator<(DrawShape &other);
      bool operator>(DrawShape &other);
      bool operator==(DrawShape &other);

      //Retrieve and convert this to the desired type, argument required for gathering the type.
      //example call DrawRectangle *rect = ourScene->getChild("id")->getThis<DrawRectangle>();
      template<typename ConvertType>
      ConvertType* getThis(){
         return (ConvertType *)this;
      }
   protected:
      //Command passing
      virtual std::string getMessage(const std::string &message){return "UNHANDLED";}
      void alertParent(const std::string &message){if(myParent!=0){myParent->getMessage(message);}}
      
      void pushMatrix();
      void popMatrix();

      Point scaleTo;
      Point translateTo;
      Point rotateTo;
      Point rotateOrigin;

      bool depthOverride;
      double overrideDepthValue;

      bool hasTexture;
      bool needsBoundRecalculate;
      GLuint *texture;
      std::vector<DrawPoint> Pnt;
      DrawShape *myParent;

      Draw2D *renderer;
   };

   class DrawPixel : public DrawShape{
      DrawPixel(){
         Pnt.resize(1);
      }
      DrawPixel(Draw2D *context):DrawShape(context){
         Pnt.resize(1);
      }
      virtual ~DrawPixel(){}

      virtual void draw();

      void setPoint(double x, double y, double z);
      void setPoint(DrawPoint point);

      template<typename PointAssign>
      void applyToPoint(PointAssign values);
   };

   template<typename PointAssign>
   void DrawPixel::applyToPoint(PointAssign values){
      alertParent("NeedsSort");
      Pnt[0] = values;
   }

   class DrawLine : public DrawShape{
      DrawLine(){
         Pnt.resize(2);
      }
      DrawLine(Draw2D *context):DrawShape(context){
         Pnt.resize(2);
      }

      virtual ~DrawLine(){}

      virtual void draw();

      void setEnds(double x1, double y1, double z1, double x2, double y2, double z2);
      void setEnds(Point startPoint, Point endPoint);

      template<typename PointAssign>
      void applyToEnds(PointAssign startPoint, PointAssign endPoint);
   };

   template<typename PointAssign>
   void DrawLine::applyToEnds(PointAssign startPoint, PointAssign endPoint){
      alertParent("NeedsSort");
      Pnt[0] = startPoint;
      Pnt[1] = endPoint;
   }

   class DrawRectangle : public DrawShape{
   public:
      DrawRectangle(){
         Pnt.resize(4);
         resetTextureCoordinates();
      }
      DrawRectangle(Draw2D *context):DrawShape(context){
         Pnt.resize(4);
         resetTextureCoordinates();
      }
      virtual ~DrawRectangle(){;}

      virtual void draw();

      void setTwoCorners(double x1, double y1, double z1, double x2, double y2, double z2);
      void setTwoCorners(Point TopLeft, Point TopRight);

      template<typename PointAssign>
      void applyToCorners(PointAssign TopLeft, PointAssign TopRight, PointAssign BottomLeft, PointAssign BottomRight);

      void resetTextureCoordinates();
   };

   template<typename PointAssign>
   void DrawRectangle::applyToCorners(PointAssign TopLeft, PointAssign TopRight, PointAssign BottomRight, PointAssign BottomLeft){
      alertParent("NeedsSort");
      Pnt[0] = TopLeft;
      Pnt[1] = BottomLeft;
      Pnt[2] = BottomRight;
      Pnt[3] = TopRight;
   }

   void AssignTextureToRectangle(DrawRectangle &rectangle, SubTexture *texture, bool flip = false);
   void AssignTextureToRectangle(DrawRectangle &rectangle, MainTexture *texture, bool flip = false);
   void AssignTextureToRectangle(DrawRectangle &rectangle, GLuint *texture, bool flip = false);

   class Scene : public DrawShape{
   public:
      Scene();
      Scene(Draw2D *context);
      virtual ~Scene();

      virtual void setRenderer(Draw2D *context);

      virtual DrawShape* add(DrawShape *childItem, const std::string &childId);
      virtual bool remove(DrawShape* childItem, bool deleteOnRemove = true);
      virtual bool remove(const std::string &childId, bool deleteOnRemove = true);

      virtual void clear();

      virtual DrawShape* getChild(const std::string& childId);

      virtual double getDepth();

      virtual BoxAABB getAABB();

      virtual void setColor(float R, float G, float B, float A);
      virtual void setColor(Color newColor);

      virtual void draw();

      //true to render the sorted list, false to render unordered.  Default behavior is true.
      void sortScene(bool depthMatters);
   protected:
      virtual std::string getMessage(const std::string &message);

      bool drawSorted, isSorted;
      std::vector<DrawShape*> DrawListVector;
      std::map<std::string, DrawShape*> DrawList;

   private:
      void sortedRender();
      void unsortedRender();
   };

   //World Scenes are kind of like the root level scene manager.
   class WorldScene {
   public:
      WorldScene(Draw2D *theRenderer){renderer = theRenderer;}
      ~WorldScene(){}

      Draw2D* getRenderer(){return renderer;}

      Scene* makeNewScene(std::string Name, bool NameIndicatesGroup = false);
      Scene* addScene(Scene* addedScene, std::string Name, bool NameIndicatesGroup = false);
      std::string getNewSceneName();

      Scene* getScene(const std::string &Name);
      bool removeScene(const std::string &Name);
      bool removeScene(Scene* delScene);

      void draw();
      void clear();

      void placeView(const Point &location){
         viewLocation = location;
      }
      void placeView(double x, double y, double z = 0){
         viewLocation.placeAt(x, y, z);
      }
      void panView(const Point &distance){
         viewLocation+=distance;
      }
      void panView(double x, double y, double z = 0){
         viewLocation.translate(x, y, z);
      }
   private:
      Draw2D* renderer;
      std::map<std::string, int> groupIdList;
      Scene rootScene;
      std::string recentName;
      Point viewLocation;
   };

}

#endif
