#include "text.h"

namespace M2Rend {

   /*************************\
   | -----TextCharacter----- |
   \*************************/

   void TextCharacter::assignSDLSurface(SDL_Surface *newSurface, char glyphChar){
      if(newSurface != nullptr){
         set = true;
         character = glyphChar;

         int widthPowerOfTwo = M2Util::roundUpPowerOfTwo(newSurface->w);
         int heightPowerOfTwo = M2Util::roundUpPowerOfTwo(newSurface->h);

         size = Point(newSurface->w, newSurface->h);
         textureSize = Point(widthPowerOfTwo, heightPowerOfTwo);

         assignSurfaceToOpenGL(newSurface, widthPowerOfTwo, heightPowerOfTwo);
      }else{
         std::cerr << "An invalid surface was assigned to a glyph." << std::endl;
      }
   }

   void TextCharacter::assignSurfaceToOpenGL( SDL_Surface *newSurface, int widthPowerOfTwo, int heightPowerOfTwo ){
      SDL_Surface *surface = SDL_CreateRGBSurface(0, widthPowerOfTwo, heightPowerOfTwo, 32, 
         M2Rend::SDL_RMASK, M2Rend::SDL_GMASK, M2Rend::SDL_BMASK, M2Rend::SDL_AMASK);

      SDL_SetAlpha(newSurface, 0, 0);
      SDL_BlitSurface(newSurface, 0, surface, 0);

      SDL_FreeSurface(newSurface);
      generateTexture(surface);
   }

   void TextCharacter::generateTexture( SDL_Surface *surface ){
      if(surface != nullptr){
         glGenTextures(1, &textureId);
         glBindTexture(GL_TEXTURE_2D, textureId);
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
         glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
         glTexImage2D(GL_TEXTURE_2D, 0, 4, surface->w, surface->h, 0, GL_BGRA, GL_UNSIGNED_BYTE, surface->pixels);
         SDL_FreeSurface(surface);
      }else{
         std::cerr << "There was a problem generating the intermediate surface for a text character." << std::endl;;
      }
   }

   bool TextCharacter::isSet(){
      return set;
   }

   char TextCharacter::getChar(){
      return character;
   }

   GLuint* TextCharacter::getTextureId(){
      return &textureId;
   }

   Point TextCharacter::getCharacterSize(){
      return size;
   }

   Point TextCharacter::getTextureSize(){
      return textureSize;
   }

   /*************************\
   | ----------Text--------- |
   \*************************/

   TextLibrary::TextLibrary( M2Rend::Draw2D *rendering ){
      white.r = 255; white.g = 255; white.b = 255; white.unused = 0;
      render = rendering;
      if(!TTF_WasInit()){
         TTF_Init();
      }
   }

   TextLibrary::~TextLibrary(){
      std::for_each(loadedFonts.begin(), loadedFonts.end(), [](std::pair<const std::string, TTF_Font*> &font){
         TTF_CloseFont(font.second);
      });
   }

   bool TextLibrary::loadFont( std::string identifier, int pointSize, std::string fontFileLocation ){
      if(loadedFonts.find(identifier) == loadedFonts.end()){
         TTF_Font* newFont;
         if(newFont = TTF_OpenFont(fontFileLocation.c_str(), pointSize)) {
            loadedFonts[identifier] = newFont;
            return true;
         }else{
            std::cerr << "Error loading font: " << TTF_GetError() << std::endl;
         }
      }else{
         std::cerr << "Error, font identifier already used, cannot assign another font to the same id: " << identifier << std::endl;
      }
      return false;
   }

   M2Rend::Scene* TextLibrary::composeScene( std::string identifier, std::string text ){
      if(loadedFonts.find(identifier) != loadedFonts.end()){
         M2Rend::Scene *textScene = new M2Rend::Scene(render);
         std::vector<TextCharacter> *characterList = initGlyphs(identifier, text);
         double characterLocation = 0;
         int characterCount = 0;
         std::for_each(text.begin(), text.end(), [&](char renderChar){
            M2Rend::DrawRectangle *character = new M2Rend::DrawRectangle(render);
            character->setTwoCorners(0, 0, 0, (*characterList)[renderChar].getTextureSize().x, (*characterList)[renderChar].getTextureSize().y, 0);
            character->translate(characterLocation, 0, 0);
            AssignTextureToRectangle(*character, (*characterList)[renderChar].getTextureId());
            textScene->add(character, boost::lexical_cast<std::string>(characterCount));
            characterLocation += (*characterList)[renderChar].getCharacterSize().x;
            characterCount++;
         });
         return textScene;
      }else{
         std::cerr << "A text scene was requested for a font identifier which is not loaded! (" << identifier << ")" << std::endl;
         return nullptr;
      }
   }

   void TextLibrary::reloadTextures(){
      std::for_each(cachedGlyphs.begin(), cachedGlyphs.end(), [&](std::pair<const std::string, std::vector<TextCharacter>> &glyphList){
         TTF_Font *currentFont = loadedFonts[glyphList.first];
         std::for_each(glyphList.second.begin(), glyphList.second.end(), [&](TextCharacter &glyph){
            if(glyph.isSet()){
               char glyphChar = glyph.getChar();
               glyph.assignSDLSurface(TTF_RenderText_Blended(currentFont, &glyphChar, white), glyphChar);
            }
         });
      });
   }

   std::vector<TextCharacter>* TextLibrary::initGlyphs( const std::string &identifier, const std::string &text ){
      std::vector<TextCharacter> *characterList = createGlyphContainer(identifier);
      loadIndividualGlyphs(identifier, text, *characterList);
      return characterList;
   }

   std::vector<TextCharacter>* TextLibrary::createGlyphContainer( const std::string &identifier ){
      if(cachedGlyphs.find(identifier) == cachedGlyphs.end()){
         cachedGlyphs[identifier] = std::vector<TextCharacter>(256);
      }
      return &(cachedGlyphs[identifier]);
   }

   void TextLibrary::loadIndividualGlyphs( const std::string &identifier, const std::string &text, std::vector<TextCharacter> &characterList ){
      TTF_Font* fontFace = loadedFonts[identifier];
      std::for_each(text.begin(), text.end(), [&](char renderChar){
         if(!characterList[renderChar].isSet()){
            characterList[renderChar].assignSDLSurface(TTF_RenderText_Blended(fontFace, &renderChar, white), renderChar);
         }
      });
   }
}
