#ifndef __TEXTURES_H__
#define __TEXTURES_H__

#include <fstream>
#include <map>
#include <vector>
#include <string>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <SDL/SDL_image.h>
#ifndef GLNULL
#define GLNULL 0
#endif
#if defined(WIN32) && !defined(GL_BGR)
#define GL_BGR GL_BGR_EXT
#endif

#ifndef WIN32
#define GLNULL GLuint(NULL)
#endif

namespace M2Rend {
/*---------------------------------------*\
| SubTexture:                             |
| Acts to store information under a main  |
| texture.  Holds a pointer to the GLuint |
| of the parent and pixel coordinates of  |
| the location on the main texture that   |
| make up the sub-texture.                |
|                                         |
| percent** holds the percent values of   |
| the main texture height that the sub    |
| texture holds (useful in application)   |
\*---------------------------------------*/
   class SubTexture{
   public:
      std::string name;
      GLuint *parentTexture;
      int leftX, topY, width, height;
      float percentLeftX, percentTopY, percentWidth, percentHeight;
      int mainHeight, mainWidth;
   };

/*---------------------------------------*\
| MainTexture:                            |
| Stores a single texture and keeps track |
| of several sub-textures.  Maintains the |
| GLuint identifier, file name, and name  |
| handle which the programmer uses to     |
| call this texture info in the future.   |
\*---------------------------------------*/

   class MainTexture{
   public:
      MainTexture(){}
      MainTexture(std::string nam, int widt, int heigh, std::string fil){
         name = nam; width = widt; height = heigh; file = fil;
         doNotSave = 0;
      }
      std::string name;
      GLuint texture;
      int width, height;
      std::string file;
      std::map<std::string, SubTexture> subTextures;
      bool doNotSave;
      bool clamped;
   };

/*---------------------------------------*\
| TextureManager:                         |
| Maintains several MainTextures and lets |
| users save the current texture info or  |
| load a previous state.  Also allows     |
| simple adding of textures and recalling |
| stored textures to apply to polygons.   |
\*---------------------------------------*/
   class TextureManager{
   public:
      TextureManager();
      ~TextureManager();
      
      bool loadTexture(const std::string &name, const std::string &file, bool repeat = 0);
      
      bool addSubTexture(const std::string &mainName, const std::string &subName, int leftX, int topY, int width, int height);
      bool addSubTexture(const std::string &mainName, SubTexture &subTextureToAdd);
      
      MainTexture *getMainTexture(const std::string &mainName);
      GLuint *getMainTextureGLuint(const std::string &mainName);
      SubTexture *getSubTexture(const std::string &mainName, const std::string &subName);
      SubTexture *getSubTexture(MainTexture &mainText, int subNum);
      SubTexture *getSubTexture(MainTexture &mainText, const std::string &subName);
      
      GLuint *getSubTextureGLuint(const std::string &mainName, const std::string &subName);
      MainTexture *getMainTexture(int num);
      SubTexture *getSubTexture(int mainNum, int subNum);
      SubTexture *getSubTexture(const std::string &mainName, int subNum);
      
      int getNumMainTextures(){return (int)mainTextures.size();}
      int getNumSubTextures(const std::string &mainName);
      int getNumSubTextures(int mainNum);
      bool deleteMainTexture(const std::string &mainName);
      bool deleteMainTexture(int num);
      bool deleteSubTexture(const std::string &mainName, const std::string &subName);
      bool deleteSubTexture(int mainNum, const std::string &subName);
      bool deleteSubTexture(const std::string &mainName, int subNum);
      bool deleteSubTexture(int mainNum, int subNum);
      void clear(){mainTextures.clear();}
      
      GLuint createEmptyTexture(const std::string &name, unsigned int width, unsigned int height);
      
      bool save(const std::string &file);
      bool load(const std::string &file);

      void reloadAllTextures(){
         std::map<std::string, MainTexture>::iterator cell;
         for(cell = mainTextures.begin();cell != mainTextures.end();cell++){
            reloadTexture(cell->second);
         }
      }
   private:
      bool reloadTexture(MainTexture &ourTexture);
      Uint32 getpixel(SDL_Surface *surface, int x, int y);
      bool loadTextureInternal(const std::string &file, GLuint &imageLoaded, int &w, int &h, bool clampIt);
      std::map<std::string, MainTexture> mainTextures;
   };
}
#endif
